"""Block logic.

Shapes are composed of many blocks while they are falling.
Once they have finished falling, they become independant.
"""


from __future__ import division

import pygame
import struct
from collections import namedtuple

from fallingsky.score import Keeper


class Blocks(object):
    rgba_codes = {  # name: (R, G, B, A)
        "black": (0, 0, 0, 255),
        "blue": (0, 87, 132, 255),
        "bonus_blue": (13, 27, 209, 180),
        "bonus_blue_alpha": (13, 27, 209, 80),
        "bonus_green": (56, 241, 6, 180),
        "bonus_green_alpha": (56, 241, 6, 80),
        "bonus_orange": (240, 120, 6, 180),
        "bonus_orange_alpha": (240, 120, 6, 80),
        "bonus_pink": (215, 8, 205, 180),
        "bonus_pink_alpha": (215, 8, 205, 80),
        "bonus_yellow": (214, 216, 8, 180),
        "bonus_yellow_alpha": (214, 216, 8, 80),
        "brown": (73, 60, 43, 255),
        "dark_orange": (164, 100, 34, 255),
        "dark_pink": (208, 60, 94, 255),
        "dark_purple": (68, 6, 110, 255),
        "dark_yellow": (180, 154, 4, 255),
        "ghost": (198, 198, 198, 120),
        "ghost_darker": (134, 134, 134, 120),
        "green": (68, 137, 26, 255),
        "light_blue": (49, 162, 242, 255),
        "light_green": (163, 206, 39, 255),
        "light_red": (219, 70, 104, 255),
        "orange": (235, 137, 49, 255),
        "pink": (224, 111, 139, 255),
        "purple": (136, 4, 118, 255),
        "red": (190, 38, 51, 255),
        "teal": (178, 220, 239, 255),
        "yellow": (247, 226, 107, 255),
    }

    Color = namedtuple("Color", ("main", "border"))
    colors = {
        "bonus_1": Color("bonus_yellow", "bonus_yellow_alpha"),
        "bonus_2": Color("bonus_green", "bonus_green_alpha"),
        "bonus_3": Color("bonus_orange", "bonus_orange_alpha"),
        "bonus_4": Color("bonus_blue", "bonus_blue_alpha"),
        "bonus_5": Color("bonus_pink", "bonus_pink_alpha"),
        "i": Color("blue", "light_blue"),
        "j": Color("pink", "dark_pink"),
        "l": Color("red", "light_red"),
        "o": Color("dark_orange", "orange"),
        "s": Color("green", "light_green"),
        "shadow": Color("ghost", "ghost_darker"),
        "t": Color("purple", "dark_purple"),
        "wall": Color("brown", "black"),
        "z": Color("yellow", "dark_yellow"),
    }

    @staticmethod
    def image_as_string(block, size):
        """Produces a block image as a RGBA string for pygame.image to load.

        Args::

            block: string block name, one of l, j, s, z, i, o, t, wall or ghost
            size: integer pixel height + width of the block

        Returns:
            RGBA as a string for the block with its colors and size
        """

        if block not in Blocks.colors.keys():
            raise TypeError("block {} is not a known block!".format(block))
        elif not hasattr(Blocks, "_image_cache"):
            setattr(Blocks, "_image_cache", {})

        if block not in Blocks._image_cache:
            Blocks._image_cache[block] = Blocks._gen_image_string(block, size)

        return Blocks._image_cache[block]

    @staticmethod
    def _gen_image_string(block, size):
        """Generates the image string for the block to then cache."""

        # packs a single pixel as a string by color name
        pixel = lambda x : struct.pack("4B", *Blocks.rgba_codes[x])

        main_color = Blocks.colors[block].main
        border_color = Blocks.colors[block].border

        # rows which are fully border pixels
        full_border = [pixel(border_color)] * size

        if block == "wall":  # special pattern
            # rows which are not patterend...
            not_patterned = [pixel(border_color)]
            not_patterned.extend([pixel(main_color)] * (size - 2))
            not_patterned.append(pixel(border_color))

            # rows which are patterend (alternating main and border)
            patterend = [pixel(border_color)]
            _pattern_color = main_color
            _next_color = border_color
            for _ in range(1, size - 3, 2):
                patterend.extend([pixel(_pattern_color)] * 2)
                _pattern_color, _next_color = _next_color, _pattern_color
            while len(patterend) < size - 1:
                patterend.append(pixel(main_color))
            patterend.append(pixel(border_color))

            # build the entire block with the above definitions as rows
            block_pattern = list(full_border)
            _current_row = not_patterned
            _next_row = patterend
            for _ in range(1, size - 3, 2):
                block_pattern.extend(_current_row * 2)
                _current_row, _next_row = _next_row, _current_row
            while (len(block_pattern) / size) < size - 1:
                block_pattern.extend(not_patterned)
            block_pattern.extend(full_border)

        elif block.startswith("bonus_"):  # bonus block patterns
            level = int(block.split("_")[1])
            quarter = int(size / 4)

            block_pattern = []
            row_start = "main"
            next_row_start = "border"
            outer = Blocks.colors[block]

            if level > 1:
                inner = Blocks.colors["bonus_{}".format(level - 1)]
            else:
                inner = outer

            for row in range(size):
                current = row_start
                next_ = next_row_start
                row_start, next_row_start = next_row_start, row_start
                ab_row = []
                for column in range(size):
                    if quarter < row < quarter * 3:
                        if quarter < column < quarter * 3:
                            ab_row.append(pixel(getattr(inner, current)))
                        else:
                            ab_row.append(pixel(getattr(outer, current)))
                    else:
                        ab_row.append(pixel(getattr(outer, current)))
                    current, next_ = next_, current
                block_pattern.append("".join(ab_row))

        else:  # normal block pattern
            # determine dot size relative to block size at 1:8 ratio
            dot = int((1 / 8) * size)

            # not patterened rows, mostly main color with borders on the side
            not_patterned = [pixel(border_color)]
            not_patterned.extend([pixel(main_color)] * (size - 2))
            not_patterned.append(pixel(border_color))

            # patterned rows being rows with the dot in them
            patterned = [pixel(border_color)]
            patterned.extend([pixel(main_color)] * (size - 2 - (dot * 2)))
            patterned.extend([pixel(border_color)] * dot)
            patterned.extend([pixel(main_color)] * dot)
            patterned.append(pixel(border_color))

            # put together the block with the above row definitions
            block_pattern = list(full_border)
            block_pattern.extend(not_patterned * (size - 2 - (dot * 2)))
            block_pattern.extend(patterned * dot)
            block_pattern.extend(not_patterned * dot)
            block_pattern.extend(full_border)

        return b"".join(block_pattern)


class Block(pygame.sprite.Sprite):
    def __init__(self, location, block, bonus_points, game, *groups):
        super(Block, self).__init__(*groups)
        self.image = pygame.image.fromstring(
            Blocks.image_as_string(block, game.blocksize),
            (game.blocksize,) * 2,
            "RGBA",
        )
        self.rect = pygame.rect.Rect(location, (game.blocksize,) * 2)
        self.exploding = False
        self.bonus_points = bonus_points
        self.visible = True
        self.name = block

    def update(self, dt, game):
        if self.exploding:
            # this sprite is dead
            self.explode(game)

    def explode(self, game):
        """Explodes this block."""

        # TODO: add animation
        self.kill()
        self.remove()
        if self.bonus_points:
            # this is a multiplier for the last score added
            game.score.game.multiply_last(self.bonus_points)
            my_location = (self.rect.x, self.rect.y)
            next_level = self.bonus_points - 1
            if next_level:
                game.spawn_bonus_block(my_location, next_level)
            else:
                game.bonus_blocks.remove(my_location)


if __name__ == "__main__":
    print(Blocks.image_as_string("bonus_4", 17))



"""User data storage and retrieval."""


from __future__ import print_function

import io
import os
import sys
import json
import appdirs


class UserData(object):
    """Saves and retrieves and is the user data.

    Initialize with the user's ID.

    usage example:

        user_data = UserData(user_id)
        play_level(user_data) # gameplay, updating user_data like a dict.. eg:
        user_data["levels_played"] += 1
        user_data.save()
    """

    def __init__(self, user_id):
        self.user_id = user_id
        self.path = os.path.join(appdirs.user_data_dir(), "Falling Sky")
        self.file = os.path.join(self.path, "{}.json".format(self.user_id))
        if not os.path.isdir(self.path):
            os.makedirs(self.path)
        self.defaults = {
            "user_id": self.user_id,
            "wins": 0,
            "losses": 0,
            "width": 10,
            "height": 25,
            "total_score": 0,
            "best_score": 0,
            "nexts": 4,
            "blocksize": 4,  # this is multiplied by 4 to ensure proper centres
            "fallrate": 1,
            "bonus_block_rate": 0,
            "show_shape_spawn_rate": False,
        }
        self.data = self._fetch()

    def _fetch(self):
        """Fetches the data from the user data store."""

        if os.path.isfile(self.file):
            try:
                with io.open(os.path.join(self.path, self.file)) as openuser:
                    return json.load(openuser)
            except Exception as error:
                print("save file corruption. removing.")
                print(error, file=sys.stderr)
                os.remove(self.file)

        return self.defaults

    def save(self):
        """Saves the user data with the user data store."""

        with open(self.file, "w") as openuser:
            json.dump(self.data, openuser)

    def __getitem__(self, item):
        """Access via self["item"] to self.data["item"]."""

        return self.data.__getitem__(item)

    def __setitem__(self, key, value):
        """Setitem's on the self.data key."""

        return self.data.__setitem__(key, value)

    def get(self, item, default=None):
        """Returns Dict.get on self.data."""

        return self.data.get(item, default)

    def update(self, other):
        """Dictionary update on self.data. THIS IS NOT THE SAVE METHOD!"""

        return self.data.update(other)



"""Generates invisible and visible walls."""


from __future__ import division


def arcade_mode(resolution=(960, 640), blocksize=16, width=20, height=70):
    """Builds regular boring, but scalable, rectangular walls.

    If given values that are too much to fit at the resolution and blocksize,
    it just fits as many as it can with some amount of space for next blocks,
    the hold block, stats...

    Args::

        resolution: (x, y) tuple of total screen width and height
        blocksize: pixel height per block
        width: integer number of blocks wide for the map
        height: integer number of blocks high for the map

    Returns::

        a tuple of a list of (x, y) coords of all walls (includes invisible),
        and a list of (x, y) coords of all visible walls for sprites/images
    """

    blocks_wide = int(resolution[0] / blocksize)
    blocks_high = int(resolution[1] / blocksize)

    mapwidth = str(blocks_wide)
    mapheight = str(blocks_high)
    tilewidth = tileheight = str(blocksize)

    centre_block = int(blocks_wide / 2)
    spread = int(width / 2)

    walls = []        # walls includes the invisible walls
    wall_blocks = []  # the coords to spawn wall blocks on
    for x_coord in range(centre_block - (spread + 1),
                         centre_block + (width - spread) + 1):
        # making walls the entire y-axis to prevent over the top/overhangs
        for y_coord in range(blocks_high):
            if (x_coord == centre_block - (spread + 1)) or \
                   (x_coord == centre_block + (width - spread)) or \
                   (y_coord == blocks_high - 2):
                # but only spawn wall block sprites where walls are :)
                coord = ((x_coord * blocksize), (y_coord * blocksize))
                if blocks_high - height < y_coord < blocks_high - 1:
                    wall_blocks.append(coord)
                walls.append(coord)

    return walls, wall_blocks


# TODO: more/different configurations



# coding: utf-8
"""Holds the MainMenu object and between-game interaction."""


from __future__ import unicode_literals

import pygame
from kezmenu import KezMenu

from fallingsky.game import GameBoard
from fallingsky.user import UserData
# from fallingsky.util import load_image


class MainMenu(object):
    """The Main menu object class which contains all children.

    Initialized with the screen resolution as an (x, y) tuple
    """

    def __init__(self, resolution, *args, **kwargs):
        self.resolution = resolution
        self.running = True
        self.injected = False
        self.updating_name = False
        self.help = False
        self.font = pygame.font.SysFont("arial", 20)
        super(MainMenu, self).__init__(*args, **kwargs)

    def run_forever(self):
        """Maybe not quite forever, but until the user quits."""

        # start the screen object
        screen = pygame.display.set_mode(self.resolution)

        self.player_name = "Player 1"  # TODO: ID auto lookup/registration
        self.data = UserData(self.player_name)

        clock = pygame.time.Clock()
        # background = load_image("background.png")

        self.menu = KezMenu(
            ["Arcade Mode", lambda: GameBoard().main(screen, self)],
            [self.player_name, self.update_player_name],
            ["Controls", lambda : setattr(self, "help", not self.help)],
            ["Quit", lambda: setattr(self, 'running', False)],
        )
        # TODO: change this. should make a player screen where they can see
        #       what accounts have been made locally, what their stats are,
        #       what their upper limits are, rename them, cheat, etc...
        PLAYER_INDEX = 1  # index of the player's name

        self.menu.x = self.resolution[0] // 3
        self.menu.y = self.resolution[1] // 2
        # self.menu.enableEffect('raise-col-padding-on-focus', enlarge_time=0.1)

        render = lambda x : self.font.render(x, True, (222, 222, 222))

        while self.running:
            events = pygame.event.get()

            if self.updating_name is True:
                events = self.handle_namechange(events)

            self.menu.options[PLAYER_INDEX]["label"] = self.player_name

            self.menu.update(events, clock.tick(30) / 1000.0)
            screen.fill((99, 99, 99))  # TODO: make better
            # screen.blit(background, (0, 0))

            if self.help:
                # add controls/help to bottom
                controls = [
                    # render("-- Controls --"),
                    render("left: ← or A"),
                    render("right: → or D"),
                    render("down: ↓ or S"),
                    render("rotate clockwise: ↑, W, or E"),
                    render("rotate counter-clockwise: Q"),
                    render("hold/swap: H"),
                    render("slam down: space bar"),
                ]

                for i, control in enumerate(controls):
                    screen.blit(control, (self.resolution[0] - 275,
                                          self.resolution[1] - 200 + (i * 25)))

            self.menu.draw(screen)
            pygame.display.flip()
            if self.data["wins"] or self.data["losses"]:
                label = "wins: {wins} losses: {losses}".format(**self.data.data)
                if not self.injected:
                    self.injected = True
                    self.menu.options.append(
                        {"label": label, "callable": lambda : 1}
                    )
                else:
                    for option in self.menu.options:
                        if option["label"].startswith("wins: "):
                            option["label"] = label
                            break

    def handle_namechange(self, events):
        """Proccess events during a name change."""

        scrubbed_events = []
        for event in events:
            if "KeyDown" in pygame.event.event_name(event.type):
                if event.key == 27:  # esc/cancel
                    self.player_name = self._old_player_name
                    self.update_player_name()
                elif event.key in (92, 8):  # backspace(s)
                    self.player_name = self.player_name[:-1]
                elif event.key not in (13,):  # 13 == enter
                    try:
                        key = chr(event.key)
                    except ValueError:
                        if event.key not in range(273, 277):
                            scrubbed_events.append(event)
                    else:
                        if pygame.key.get_mods() in (1, 8192):
                            # checks for shift or caps lock, and both
                            key = key.capitalize()
                        self.player_name += key
                else:
                    scrubbed_events.append(event)
            else:
                scrubbed_events.append(event)

        # 2nd item in the list is the players name, update it
        return scrubbed_events

    def update_player_name(self):
        """Called when someone clicks on the player name."""

        if self.updating_name:
            self.updating_name = False
            if not self.player_name:  # pressing enter twice, no input, cancels
                self.player_name = self._old_player_name
                return
            self._old_player_name = None
            self.menu.mouse_enabled = True
            self.data = UserData(self.player_name)
            if self.injected:
                self.menu.options.pop(-1)
                self.injected = False
        else:
            self.updating_name = True
            self._old_player_name = self.player_name  # save this for cancel
            self.player_name = ""
            self.menu.mouse_enabled = False



"""Launcher scripts."""


import pygame

from fallingsky import shapes
from fallingsky.menu import MainMenu
from fallingsky.game import GameBoard


# 3:2, 960x640 -- shared resolution between android and ios
STANDARD_RESOLUTION = (960, 640)


def play():
    """Main access point, plays game through the main menu."""

    pygame.init()
    menu = MainMenu(STANDARD_RESOLUTION)
    try:
        menu.run_forever()
    except KeyboardInterrupt:
        raise SystemExit("Interrupted")


def play_hack():
    """Quick and dirty dev/testing/cheating access, skips the main menu."""

    class HackMenu(object):
        data = {             # feel free to edit to your liking
            "user_id": "hacker",
            "wins": 0,
            "losses": 0,
            "width": 10,     # number of blocksize units of gameboard width
            "height": 25,    # number of blocksize units of gameboard height
            "total_score": 0,
            "best_score": 0,
            "nexts": 4,      # how many next blocks you get
            "blocksize": 4,  # this is multiplied by 4 to ensure proper centres
            "fallrate": 1,   # drop speed, aka level. from 1-21
            "bonus_block_rate": 0,  # number of bonus blocks/number of wins
            "show_shape_spawn_rate": False,  # shows your % shape spawns
        }
        def __init__(self, res):
            self.resolution = res
            self.losses = 0
            self.wins = 0

    # uncomment the next line if you want every shape to be a line
    # shapes.IMADEVELOPER = True

    GameBoard().main(
        pygame.display.set_mode(STANDARD_RESOLUTION),
        HackMenu(STANDARD_RESOLUTION),
    )


if __name__ == "__main__":
    # to hack, run this file directly instead of the fallingsky command
    pygame.init()
    play_hack()



"""The Tragedy of the Falling Sky's main game loop."""


from __future__ import division

import os
import pygame
import random
import tempfile

from collections import Counter
from collections import namedtuple
from collections import defaultdict

from fallingsky import __version__
from fallingsky.block import Block
from fallingsky.score import Keeper
from fallingsky.shapes import Shape
from fallingsky.shapes import Shapes
from fallingsky.user import UserData
from fallingsky.util import load_image
# from fallingsky.util import load_sound
from fallingsky.walls import arcade_mode


class GameBoard(object):
    """Main game object, passed to other classes instantiated as `game`.

    Controls main logic for a level of gameplay, spawning Sprite classes.

    Initialized with no arguments. (init isn't important)

    GameBoard.main() is the method called per level, which interacts with the
    game menu object defined further below.
    """

    def __init__(self):
        self.fonts = {
            "normal": pygame.font.SysFont("arial", 28),
            "small": pygame.font.SysFont("courier new", 8),
        }
        self.score = Keeper()
        self.lines = 0

    def render(self, text, font="normal", color=None):
        """Uses pygame's font.render to make some text.

        Args::

            test: string text to render
            font: string key font name in self.fonts
            color: a tree integer tuple RGB color code, or white

        Returns:
            font object to blit somewhere
        """

        return self.fonts[font].render(text, True, color or (255, 255, 255))

    def refresh_background(self, dt):
        """Called per clock cycle, updates all graphics."""

        self.screen.fill((111, 111, 111))  # TODO: make better

        # go through the game sprites and update/reblit them if visible
        for sprite in self.sprites:
            sprite.update(dt, self)
            if sprite.visible:
                self.screen.blit(sprite.image, (sprite.rect.x, sprite.rect.y))

        # grab some current game stats
        stats = [
            self.render("level: {}".format(self.fallrate)),
            self.render("lines: {:,}".format(self.lines)),
            self.render("game: {:,}".format(int(self.score.game))),
        ]

        if self.score.total:  # only after the first loss
            stats.append(self.render("total: {:,}".format(self.score.total)))

        if self.score.best != self.score.total:  # after 2nd loss
            stats.append(self.render(
                "best: {:,}".format(int(self.score.best))
            ))

        if self.bonus_blocks:  # once they get 100k points
            stats.append(self.render("bonus: {:,}".format(sum([
                self.blocks[coord].bonus_points for coord in self.bonus_blocks
            ]))))
        elif self.bonus_block_rate:
            stats.append(self.render(
                "{} complete!".format(self.bonus_block_rate)
            ))
        elif self.score.game.get_score() > 100000:
            stats.append(self.render("!"))  # just a small notification ...

        if self.show_shape_spawn_rate:  # debug/info option
            shapes_spawned = sum(self.history.values())
            shape_spawn_rate = lambda x : self.history[x] / shapes_spawned
            stats.append(self.render(""))  # spacer
            stats.extend([self.render("{}: {:.2%}".format(
                Shapes.all_types[shape].title(),
                shape_spawn_rate(shape)
            )) for shape in self.history])

        for i, stat in enumerate(stats):  # draw the rendered fonts on screen
            self.screen.blit(stat, (60, 30 + (i * 30)))

        # display name and version
        self.screen.blit(
            self.render(
                "The Tragedy of the Falling Sky v{}".format(__version__),
                font="small",
                color=(0, 0, 0),
            ),
            (5, self.resolution[1] - 10),  # bottom left corner
        )

        pygame.display.flip()   # flip and we're done for this update

    def reset_blocks(self):
        """Resets self.blocks to a dict of {coords: None} for the walls."""

        self.blocks = {coord: None for coord in self.walls}
        for sprite in self.sprites:
            if hasattr(sprite, "explode"):
                sprite.explode(self)

        # respawn the walls
        for coord in self.wall_coords:
            Block(coord, "wall", 0, self, self.sprites)

        # shape history, used by shapes when spawning
        self.history = Counter({key: 0 for key in Shapes.all_types.keys()})
        self.spawn_bonus_blocks()

    def explode_full_lines(self):
        """Explodes all fully filled lines. Updates self.lines and game score.

        Returns:
            Integer count of the number of lines destroyed.
        """

        blocks_per_line = Counter()
        blocks_by_line = defaultdict(list)

        for coords, block in self.blocks.items():
            if block is not None:
                blocks_per_line[coords[1]] += 1
                blocks_by_line[coords[1]].append(block)

        destroyed_lines = []
        for line, count in blocks_per_line.items():
            if count >= self.width:
                destroyed_lines.append(line)

        # add the score /before/ exploding any blocks so the multipliers work
        num_lines_destroyed = len(destroyed_lines)
        points = int(
            ((num_lines_destroyed ** 1.5) * (100 * self.width) // 500) * 500
        )
        self.score.game += points

        # track lines, adjust the fallrate maybe
        self.lines += num_lines_destroyed
        self.lines_until_speed_up -= num_lines_destroyed
        if self.lines_until_speed_up <= 0 and self.fallrate < self.max_level:
            self.fallrate += 1
            self.lines_until_speed_up = self.lines_per_level

        # explode all the full lines, remove them from self.blocks
        for line in destroyed_lines:
            for block in blocks_by_line[line]:
                self.blocks.pop((block.rect.x, block.rect.y))
                block.explode(self)

        destroyed_lines = sorted(destroyed_lines)

        if destroyed_lines:  # this call is expensive and recursive
            num_lines_destroyed += self.blocks_fall_down(destroyed_lines)

        return num_lines_destroyed

    def _get_blocks_to_move(self, destroyed_lines):
        """Determines the lines connected to the destroyed lines.

        Returns:
            dictionary of line num: list of moved block sprites in that line
        """

        connected_lines = list(destroyed_lines)
        moved_blocks = defaultdict(list)

        line = connected_lines[0]
        line_found = True
        while line_found:  # check up
            line -= self.blocksize

            if line in connected_lines:
                continue

            line_found = False
            for coords, block in self.blocks.items():
                if coords[1] == line and block:
                    if line not in connected_lines:
                        connected_lines.append(line)
                        line_found = True
                    moved_blocks[line].append((coords, block))

        return moved_blocks

    def blocks_fall_down(self, destroyed_lines):
        """Moves the rest of the board downwards for the destroyed lines."""

        moved_blocks = self._get_blocks_to_move(destroyed_lines)
        colliders = list(self.bonus_blocks)

        for _ in range(len(destroyed_lines)):
            moved_this_loop = defaultdict(list)

            for line, blocks in reversed(sorted(moved_blocks.items())):
                for coord, block in blocks:
                    if block is None or block.bonus_points:
                        moved_this_loop[line].append((coord, block))
                    else:
                        # clean up current position if we're moving down multiple lines
                        if coord in colliders:
                            colliders.remove(coord)
                        desired_move = (coord[0], coord[1] + self.blocksize)
                        if desired_move in colliders:
                            colliders.append(coord)
                        else:
                            colliders.append(desired_move)
                            self.blocks.pop(coord, None)
                            block.rect.y += self.blocksize
                            self.blocks[(block.rect.x, block.rect.y)] = block

                        moved_this_loop[line].append((
                            (block.rect.x, block.rect.y), block
                        ))

            moved_blocks = moved_this_loop

        # it is possible that we've moved down to make another full line
        return self.explode_full_lines()

    def spawn_bonus_blocks(self):
        """Spawns bonus blocks inside the game grid."""

        self.bonus_blocks = []

        # give them some room up top (still low odds to spawn here)
        max_spawn_height = self.height - 2

        # this is pretty much standard deviation, but with /no/ outliers
        odds = {"big": 342, "mid": 137, "sml": 21}  # HACK: keys are important
        range_distribution = {key: 0 for key in odds}

        # count bonus block spawns above and below the mean
        row_distribution = {key: Counter(above=0, below=0) for key in odds}

        # mean is at a ratio of 1:3 compared to the max height
        mean = max_spawn_height // 3

        # the standard deviation length is at a ratio of 1:7 of the max height
        std = max_spawn_height // 7

        rows_above = lambda x : list(range(
            max(mean - (std * x), 0),
            max(mean - (std * (x - 1)), 0)
        ))
        rows_below = lambda x : list(range(
            min(mean + (std * x) - std, max_spawn_height),
            min(mean + (std * (x + 1)) - std, max_spawn_height)
        ))
        Rows = namedtuple("Rows", ("above", "below"))
        row_range = lambda x : Rows(below=rows_below(x), above=rows_above(x))
        # HACK: using the key names sorted here, don't change odds' keys ;)
        rows = {key: row_range(i) for i, key in enumerate(sorted(odds), 1)}

        # fucky int: int dict to track what rows to spawn bonus blocks in
        row_spawns = {row: 0 for row in range(max_spawn_height)}

        # track each column in each row for spawning inside of
        row_columns = {
            row: list(range(-int(self.width / 2), int(self.width / 2))) for \
                row in range(max_spawn_height)
        }

        for i, block in enumerate(range(self.bonus_block_rate)):

            rolls_so_far = sum(range_distribution.values())
            while True:
                # roll and determine the odds section it landed in
                roll = random.randint(0, 1000)
                if roll < odds["sml"] * 2:
                    section = "sml"
                    if range_distribution["sml"] and rolls_so_far < 10:
                        continue  # getting this way too often, shortcut
                elif roll < odds["mid"] * 2:
                    section = "mid"
                else:
                    section = "big"

                # apply some logic to smooth out the distribution
                if rolls_so_far > 2 and range_distribution[section] and \
                        range_distribution[section] / (rolls_so_far + 1) > \
                        ((odds[section] + ((1/5) * odds[section])) * 2) / 1000:
                    continue
                else:
                    break

            range_distribution[section] += 1

            above = random.randint(0, 1)

            # move above/below into the lesser populated area
            dist = row_distribution[section]
            if (dist["above"] and above and dist["above"] > dist["below"]) or \
               (dist["below"] and not above and dist["below"] > dist["above"]):
                above = 0 if above else 1

            def available_rows():
                return [row for row in rows[section][above] if \
                        row_spawns[row] < self.width - 1]

            this_row = None
            try:
                this_row = random.sample(available_rows(), 1)[0]
            except ValueError:
                above = 0 if above else 1  # swap value
                try:
                    this_row = random.sample(available_rows(), 1)[0]
                except ValueError:
                    # so many bonus blocks we've filled the distribution area
                    # ignored, but accept it statistically to spawn other areas
                    pass
            finally:
                row_spawns[this_row] += 1

            if above:
                row_distribution[section]["above"] += 1
            else:
                row_distribution[section]["below"] += 1

            column = random.sample(row_columns[this_row], 1)[0]
            row_columns[this_row].remove(column)
            level = random.randint(3, 5)
            location = (
                self.centre_px + (column * self.blocksize),
                self.vertical_offset + ((self.height - this_row - 3) * self.blocksize)
            )
            self.spawn_bonus_block(location, level)

    def spawn_bonus_block(self, location, level):
        """Creates a bonus block at location and bonus level."""

        self.blocks[location] = Block(
            location,
            "bonus_{}".format(level),
            level,
            self,
            self.sprites,
        )
        if location not in self.bonus_blocks:
            self.bonus_blocks.append(location)

    def reset_game_board(self):
        """Explodes everything and resets the gameboard."""

        # explode the board
        for _, block in self.blocks.items():
            if block is not None:
                block.bonus_points = 0
                block.explode(self)

        # explode the next queue
        for shape in self.next_queue:
            for block in shape.blocks:
                block.explode(self)

        # reset let's play again!
        # TODO: add screens/gameplay here
        self.score.game_over()
        self.reset_blocks()
        self.lines = 0
        self.next_queue = [
            Shape(game=self, position=i) for i in range(1, self.nexts)
        ]
        self.active = True
        self.fallrate = self.starting_fallrate

    def get_next_shape(self):
        """Gets the next-in-line shape from self.next_queue.

        Moves forward all other members and spawns a new one at the end.
        """

        new_shape = self.next_queue.pop(0)
        for next_shape in self.next_queue:
            next_shape.move_closer(self)
        self.next_queue.append(
            Shape(game=self, position=self.nexts - 1)
        )
        return new_shape

    def main(self, screen, menu):
        """Main Game routine. Make fun now! :D

        Args::

            screen: the PyGame screen object we're running inside
            menu: MainMenu object which called us
        """

        # grab a clock so we can limit and measure the passing of time
        clock = pygame.time.Clock()

        # TODO: make a real background
        # background = load_image("background.png")

        self.screen = screen
        self.resolution = menu.resolution
        self.blocksize = min(menu.data["blocksize"], 12) * 4

        # board size
        max_width = int(menu.resolution[0] / self.blocksize) - 10
        max_height = int(menu.resolution[1] / self.blocksize) - 3
        self.width = min(menu.data["width"], max_width)
        self.height = min(menu.data["height"], max_height)

        # speed
        self.fallrate = menu.data["fallrate"]
        self.starting_fallrate = self.fallrate

        # score
        self.score.resume(menu.data["total_score"], menu.data["best_score"])
        self.bonus_block_rate = menu.data["bonus_block_rate"]

        # "features"
        self.nexts = menu.data["nexts"] + 1  # plus 1 because of range usage
        self.show_shape_spawn_rate = menu.data["show_shape_spawn_rate"]

        # do geometry, generate the xml map
        self.centre_px = int(
            ((self.resolution[0] / self.blocksize) / 2) * self.blocksize
        )
        self.vertical_offset = int(
            ((self.resolution[1] // self.blocksize) - self.height) * self.blocksize
        )
        self.walls, self.wall_coords = arcade_mode(
            resolution=self.resolution,
            blocksize = self.blocksize,
            width=self.width,
            height=self.height,
        )

        # create a SpriteLayer for all our sprites to live in
        self.sprites = pygame.sprite.AbstractGroup()

        # spawn the walls, reset blocks
        self.reset_blocks()

        # this board will remain active until a Shape class toggles it False
        # when it finds it collides when placing itself on the board
        self.active = True

        # build initial shapes here
        current_shape = Shape(game=self)

        # TODO: display next shapes as per skill level
        # just by ID though, don't spawn them as Shape objects, or change Shape
        # init to maybe display into a "next" box of varying length
        self.next_queue = [
            Shape(game=self, position=i) for i in range(1, self.nexts)
        ]
        swappped = False
        swapped_shape = None
        # TODO: move these constants somewhere common
        self.max_level = 21
        self.lines_per_level = 8 # 16  # XXX
        self.lines_until_speed_up = self.lines_per_level
        slam_delay = 200
        slam_available = slam_delay
        swap_delay = 400
        swap_available = swap_delay

        while True:
            # limit updates to 30 times per second and determine how much time
            # passed since the last update
            dt = clock.tick(60)
            slam_available -= dt
            swap_available -= dt
            # handle basic game events; terminate this main loop if the window
            # is closed or the escape key is pressed

            keys = [i for i, key in enumerate(pygame.key.get_pressed()) if key]

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    return
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        return

            nuke_keys = False
            for key in keys:
                # swap
                if key in (pygame.K_x, pygame.K_h) and swap_available < 0 and \
                        not swappped and current_shape.falling:
                    swap_available = swap_delay
                    current_shape.become_held(self)
                    current_shape, swapped_shape = swapped_shape, current_shape
                    nuke_keys = True

                    if current_shape is None:
                        # first move to the hold area, grab next in line
                        current_shape = self.get_next_shape()
                    else:
                        swappped = True

                    if current_shape.can_activate(self):
                        current_shape.make_active(self)

                # slam
                elif key in (pygame.K_SPACE,) and slam_available < 0:
                    current_shape.slam_blocks(self)
                    nuke_keys = True

            if self.active:
                current_shape.update(dt, self, [] if nuke_keys else keys)

            self.refresh_background(dt) #, background)

            if self.active is not True:  # changes from True to int win/loss
                game_score = self.score.game.get_score()
                if self.active > 0 or \
                        (not self.bonus_blocks and self.bonus_block_rate) or \
                        (not self.bonus_block_rate and game_score > 50000) or \
                        game_score > 100000:
                    menu.data["wins"] += 1
                    self.bonus_block_rate += 1
                else:
                    menu.data["losses"] += 1

                menu.data["total_score"] += game_score
                menu.data["best_score"] = max(menu.data["best_score"], game_score)
                menu.data.save()

                self.reset_game_board()
                current_shape = self.get_next_shape()
                current_shape.make_active(self)
                if swapped_shape:
                    for block in swapped_shape.blocks:
                        block.explode(self)
                    del swapped_shape
                    swapped_shape = None
                slam_available = slam_delay
                swap_available = swap_delay
                continue

            if not current_shape.falling:  # the current shape stopped falling
                del current_shape
                destroyed_lines = self.explode_full_lines()
                if destroyed_lines:
                    # print("you destroyed {} lines!!!".format(destroyed_lines))
                    pass  # play a sound here
                current_shape = self.get_next_shape()
                current_shape.make_active(self)
                swappped = False
                slam_available = slam_delay
                swap_available = swap_delay


class SoundEffects(object):
    """TODO: add sound :)"""

    pass



import pytest

from fallingsky.user import UserData


def test_object_item_access():
    """Ensure we can set and retrieve user data as dict keys."""

    user_data = UserData(1)
    user_data["test_key"] = True
    assert user_data["test_key"] is True
    user_data["test_key"] = ["one", "two", "three"]
    assert user_data.get("test_key") == ["one", "two", "three"]
    assert user_data.get("ToTaLLy_PR0B4blY_N07_R34L") is None
    assert user_data.get("ToTaLLy_PR0B4blY_N07_R34L", "kay") == "kay"



"""Shape object for The Tragedy of the Falling Sky.

Movement handling to the child blocks happens here.
"""


from __future__ import division

import sys
import pygame
import random

from fallingsky.block import Block


IMADEVELOPER = False


class Shapes(object):
    """A lazymans enum to represent the block types and their IDs."""

    all_types = {
        0: "l",
        1: "s",
        2: "z",
        3: "i",
        4: "o",
        5: "j",
        6: "t",
    }

    def __init__(self):
        for id_, type_ in Shapes.all_types.items():
            setattr(self, type_, id_)

    @staticmethod
    def get_type(block_type_int):
        return Shapes.all_types.get(block_type_int)


class Shape(object):
    """Creates a shape as per the BlockTypes enum.

    Init args::

        game: the GameBoard object
        centre_px: integer pixel that is the centre of the window
        shape: integer shape ID to force a shape, or None for random selection
        blocksize: integer block width and height in pixels
        groups
    """

    def __init__(self, game, position=0, shape=None):
        if shape is None:
            self.shape = 3 if IMADEVELOPER else not_so_random_shape(game)
        else:
            self.shape = shape

        game.history[self.shape] += 1

        self.shape_name = Shapes.get_type(self.shape)
        self.position = position

        bs = game.blocksize
        offset_coords = {
            "l": [(-(bs * 2), 0), (-bs, 0), (0, 0), (0, -bs)],
            "j": [(-bs, -bs), (-bs, 0), (0, 0), (bs, 0)],
            "s": [(bs, -bs), (0, -bs), (0, 0), (-bs, 0)],
            "z": [(-bs, -bs), (0, -bs), (0, 0), (bs, 0)],
            "t": [(0, -bs), (-bs, 0), (0, 0), (bs, 0)],
            "o": [(-bs, -bs), (0, -bs), (0, 0), (-bs, 0)],
            "i": [(-(bs * 2), 0), (-bs, 0), (0, 0), (bs, 0)],
        }

        self.offset_coords = offset_coords[self.shape_name]
        self.vertical_offset = game.vertical_offset
        self.exploding = False
        self.falling = True
        self.fall_rate = max(1100 - (game.fallrate * 50), 65)
        self.next_fall = self.fall_rate
        self.move_rate = 100
        self.next_move = self.move_rate
        self.turn_rate = 200
        self.next_turn = self.turn_rate
        self.down_rate = 100  # need to throttle down to encourage slams
        self.down_available = self.down_rate
        self.blocks = []
        self.bottom_mercy = 1  # mercy/grace ticks for bottom stickyness

        if self.position == 0:
            # this can return False, but we don't really care right now
            block_positions = self.can_activate(game) or []
        else:
            block_positions = self._locations_in_queue(game)

        for location in block_positions:
            self.blocks.append(
                Block(location, self.shape_name, 0, game, game.sprites)
            )

        self.shadow_blocks = []

        # spawn our shawdows if we're spawning as an active shape
        if self.position == 0 and block_positions:
            self.spawn_shadow_blocks(game)

    def spawn_shadow_blocks(self, game):
        """Builds the coords for shadows based off of self.blocks.

        Args:
            game: the GameBoard object

        Returns:
            None, appends Block objects of shadow type to self.shadow_blocks
        """

        for coord in self._shadow_coords(game):
            self.shadow_blocks.append(
                Block(coord, "shadow", 0, game, game.sprites)
            )

    def _shadow_coords(self, game):
        """Determine the coords for the shadow of this shape.

        The shadow is always directly downwards from the shape.

        Returns:
            list of coords in order of self.shadow_blocks
        """

        # self.blocks exists at this point, is active, and can abuse the coords
        shadow_blocks = [(block.rect.x, block.rect.y) for block in self.blocks]
        think_of_the_bits = game.height * 2
        finished = False
        while not finished and think_of_the_bits > 0:
            think_of_the_bits -= 1
            shadow_blocks_next = []
            for block in shadow_blocks:
                new_coord = (block[0], block[1] + game.blocksize)
                if new_coord in game.blocks:
                    finished = True  # double break
                    break
                else:
                    shadow_blocks_next.append(new_coord)
            else:
                shadow_blocks = shadow_blocks_next

        return shadow_blocks if think_of_the_bits > 0 else []

    def move_closer(self, game):
        """Move to the next closest position in the next waiting area."""

        self.position -= 1
        for coord, block in zip(self._locations_in_queue(game), self.blocks):
            block.rect.x = coord[0]
            block.rect.y = coord[1]

    def _locations_in_queue(self, game):
        """Returns the block coords for the integer position in the queue.

        Args::

            game: the GameBoard object spawning us
            position: an integer position (1-10?) of our place in queue
        """

        # TODO: rows & better placement
        block_positions = []
        right_shift = (((game.width // 2) + 4) * game.blocksize)
        down_shift = (((self.position - 1) * 3) * game.blocksize)
        for block_offset in self.offset_coords:
            location = (
                game.centre_px + block_offset[0] + right_shift,
                self.vertical_offset + block_offset[1] + down_shift,
            )
            block_positions.append(location)

        return block_positions

    def become_held(self, game):
        """Move this shape to the hold area."""

        block_positions = []
        right_shift = (((game.width // 2) + 4) * game.blocksize)
        for block_offset in self.offset_coords:
            location = (
                game.centre_px + block_offset[0] - right_shift,
                self.vertical_offset + block_offset[1],
            )
            block_positions.append(location)

        for coord, block in zip(block_positions, self.blocks):
            block.rect.x = coord[0]
            block.rect.y = coord[1]

        for block in self.shadow_blocks:
            block.explode(game)
        self.shadow_blocks = []
        self.falling = False

    def make_active(self, game):
        """Move this shape from the next or hold areas to the active area.

        Args:
            game: the GameBoard object spawning us

        Returns:
            boolean of succesfully becoming active
        """

        move_locations = self.can_activate(game)
        if move_locations:
            for coords, block in zip(move_locations, self.blocks):
                block.rect.x = coords[0]
                block.rect.y = coords[1]
            self.spawn_shadow_blocks(game)
            self.falling = True
            self.fall_rate = max(1100 - (game.fallrate * 50), 65)
            self.next_fall = self.fall_rate
            return True
        else:
            return False

    def can_activate(self, game):
        """Checks if we can become the active block. Can set game.active = 0.

        Returns:
            list of positions to move self.blocks to or False if not possible
        """

        move_locations = []
        for block_offset in self.offset_coords:
            location = (
                game.centre_px + block_offset[0],
                self.vertical_offset + block_offset[1],
            )
            if location in game.blocks:
                game.active = 0
                self.falling = False
                return False
            else:
                move_locations.append(location)

        return move_locations

    def _move_coords(self, game, coords, left=False, right=False, down=False):
        """Determins the coordinates possible to move from coords in direction.

        Args::

            game: GameBoard object which requested this Shape
            coords: tuple of (x, y) coordinates currently at
            left: boolean to request leftbound movement
            right: boolean to request rightward movement
            down: boolean to request movement downwards

        Returns:
            tuple of coords (x, y), updated for the move if no collision
        """

        if left:
            desired_coords = (coords[0] - game.blocksize, coords[1])
        elif right:
            desired_coords = (coords[0] + game.blocksize, coords[1])
        elif down:
            desired_coords = (coords[0], coords[1] + game.blocksize)

        return coords if desired_coords in game.blocks else desired_coords

    def _move_blocks(self, game, left=False, right=False, down=False):
        """Moves all the blocks in self.blocks the direction asked."""

        block_moves = []
        for block in self.blocks:
            coords = (block.rect.x, block.rect.y)
            new_coords = self._move_coords(game, coords, left, right, down)
            if coords == new_coords:
                if down:
                    if self.bottom_mercy > 0:  # mercy granted, this time
                        self.bottom_mercy -= 1
                        return False

                    # we've hit ground, add coords and block to game.blocks
                    for block, shadow in zip(self.blocks, self.shadow_blocks):
                        game.blocks[(block.rect.x, block.rect.y)] = block
                        shadow.explode(game)

                    self.shadow_blocks = []
                    self.falling = False
                return False  # cancels the move, new_coords are not new
            else:
                block_moves.append(new_coords)

        # move the regular blocks
        for block, new_coords in zip(self.blocks, block_moves):
            block.rect.x = new_coords[0]
            block.rect.y = new_coords[1]

        self._update_shadow_positions(game)
        return True

    def _update_shadow_positions(self, game):
        """Moves the shadow blocks after a block move or rotate."""

        for block, coord in zip(self.shadow_blocks, self._shadow_coords(game)):
            block.rect.x = coord[0]
            block.rect.y = coord[1]

    def slam_blocks(self, game):
        """Moves all blocks as far down as possible."""

        for shadow in self.shadow_blocks:
            shadow.explode(game)
        self.shadow_blocks = []

        # we may be unable to move downwards, when slamming while losing
        for block in self.blocks:
            coords = (block.rect.x, block.rect.y)
            if self._move_coords(game, coords, down=True) == coords:
                # we're done moving, make us colliadable
                for blk in self.blocks:
                    game.blocks[(blk.rect.x, blk.rect.y)] = blk
                shadow_coords = []
                break
        else:
            shadow_coords = self._shadow_coords(game)

        if len(shadow_coords) == 4:
            for block, coords in zip(self.blocks, shadow_coords):
                block.rect.x, block.rect.y = coords[0], coords[1]
                game.blocks[coords] = block

        self.falling = False

    def _rotate_blocks(self, game, clockwise=True, retry=0):
        """Rotate the blocks in the shape in a direction if possible.

        Args:
            game: the GameBoard object which spawned this Shape object
            clockwise: boolean to rotate clockwise or counter-clockwise
            retry: boolean of if this function is being recalled after shifting

        Returns:
            boolean of if the block rotated
        """

        if self.shape_name == "o":
            return False  # square blocks don't rotate...

        requested_movements = []
        centre = (self.blocks[2].rect.x, self.blocks[2].rect.y)
        for offset, block in zip(self.offset_coords, self.blocks):
            if clockwise:
                move_offset = ((-1 * offset[1]), offset[0])
            else:
                move_offset = (offset[1], (-1 * offset[0]))

            requested_move = (
                centre[0] + move_offset[0],
                centre[1] + move_offset[1],
            )
            if requested_move in game.blocks:
                if game.blocks[requested_move] is None and retry <= 2:
                    return self._shift_retry_rotate(game, clockwise, retry)
                else:
                    return False
            else:
                requested_movements.append(requested_move)

        for requested_move, block in zip(requested_movements, self.blocks):
            block.rect.x = requested_move[0]
            block.rect.y = requested_move[1]

        self._rotate_offsets(clockwise)
        self._update_shadow_positions(game)
        return True

    def _block_locations(self):
        """Returns the current block locations in a list of coord tuples."""

        return [(block.rect.x, block.rect.y) for block in self.blocks]

    def _shift_retry_rotate(self, game, clockwise, retry=0):
        """Shift the blocks away from the walls and try to rotate again."""

        retry += 1
        start_coords = self._block_locations()

        if start_coords[0][0] > game.centre_px:  # first block, x coord
            moved = self._move_blocks(game, left=True)
        else:
            moved = self._move_blocks(game, right=True)

        rotated = False
        if moved:
            rotated = self._rotate_blocks(game, clockwise, retry)

        if rotated:
            return True

        else:
            for block, coord in zip(self.blocks, start_coords):
                block.rect.x = coord[0]
                block.rect.y = coord[1]
            return False

    def _rotate_offsets(self, clockwise=True):
        """Updates self.offsets with a rotation factor."""

        new_offsets = []
        for offset in self.offset_coords:
            if clockwise:
                new_offsets.append(((-1 * offset[1]), offset[0]))
            else:
                new_offsets.append((offset[1], (-1 * offset[0])))
        self.offset_coords = new_offsets

    def update(self, dt, game, key_presses):
        """Called per clock tick. Controls user input for shape movement."""

        if self.falling:

            self.next_fall -= dt
            self.next_move -= dt
            self.next_turn -= dt
            self.down_available -= dt

            for key in key_presses:
                self.handle_key_press(game, key)

            if self.next_fall < 0:
                self._move_blocks(game, down=True)
                self.next_fall = self.fall_rate

            # show or hide shadow blocks if we're colliding with them
            block_locations = self._block_locations()
            for shadow in self.shadow_blocks:
                if (shadow.rect.x, shadow.rect.y) in block_locations:
                    shadow.visible = False
                else:
                    shadow.visible = True

    def handle_key_press(self, game, key):
        """Logic per key press.

        Returns:
            boolean of if the shadow blocks need to be updated or not
        """

        # TODO: make a player-friendly mode to reset mercy on every move/turn

        # falling sped up
        if key in (pygame.K_DOWN, pygame.K_s) and self.down_available <= 0:
            self._move_blocks(game, down=True)
            self.down_available = self.down_rate
            self.next_fall = self.fall_rate
            return False

        # right and left
        elif key in (pygame.K_LEFT, pygame.K_a) and self.next_move <= 0:
            self._move_blocks(game, left=True)
            self.next_move = self.move_rate
        elif key in (pygame.K_RIGHT, pygame.K_d) and self.next_move <= 0:
            self._move_blocks(game, right=True)
            self.next_move = self.move_rate

        # rotate, both directions
        elif key in (pygame.K_UP, pygame.K_e, pygame.K_w) and self.next_turn <= 0:
            self._rotate_blocks(game)
            self.next_turn = self.turn_rate
        elif key in (pygame.K_q,) and self.next_turn <= 0:
            self._rotate_blocks(game, clockwise=False)
            self.next_turn = self.turn_rate


def not_so_random_shape(game):
    """Used to determine the next shape ID. Tries to weigh the odds evenly."""

    roll_for_shape = lambda : random.sample(game.history.keys(), 1)[0]
    shape_id = roll_for_shape()
    shapes_spawned = sum(game.history.values())

    if shapes_spawned > 4:
        shape_spawn_rate = lambda x : game.history[x] / shapes_spawned
        while shape_spawn_rate(shape_id) > 1 / (len(game.history) - 1):
            shape_id = roll_for_shape()

    return shape_id



"""Utility functions used in The Tragedy of the Falling Sky."""


import os
import pygame


def _here():
    """Returns the current full directory path."""

    return os.path.dirname(os.path.realpath(__file__))


def load_image(image_name):
    """Loads the image_name from the "images" directory.

    Args:
        image_name: string image file name, with extension

    Returns:
        the result of pygame.image.load for the image file
    """

    return pygame.image.load(os.path.join(_here(), "images", image_name))


def load_sound(sound_file):
    """Loads the sound_file from the "sounds" directory.

    Args:
        sound_file: string name of the sound file with extension

    Returns:
        instantiated pygame.mixer.Sound object for the sound file
    """

    return pygame.mixer.Sound(os.path.join(_here(), "sounds", sound_file))



"""The Tragedy of the Falling Sky's setup.py."""


try:
    import pygame  # have this setup beforehand
except ImportError:
    raise SystemExit("unable to import pygame.")


from setuptools import setup
from setuptools import find_packages
from setuptools.command.test import test as TestCommand


with open("fallingsky/__init__.py", "r") as openinit:
    for line in openinit.readlines():
        if line.startswith("__version__ ="):
            __version__ = line[14:].replace('"', "").replace('"', "").strip()
            break
    else:
        __version__ = "0.0.0"


class PyTest(TestCommand):
    """Shim in pytest to be able to use it with setup.py test."""

    def finalize_options(self):
        """Stolen from http://pytest.org/latest/goodpractises.html."""

        TestCommand.finalize_options(self)
        self.test_args = ["-v", "-rf", "--cov-report", "term-missing", "--cov",
                          "fallingsky", "test"]
        self.test_suite = True

    def run_tests(self):
        """Also shamelessly stolen."""

        # have to import here, outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        raise SystemExit(errno)


setup(
    name="fallingsky",
    version=__version__,
    author="Adam Talsma",
    author_email="adam@talsma.ca",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["kezmenu", "appdirs"],
    dependency_links=[
        "https://gitlab.com/a-tal/kezmenu/repository/archive.tar.bz2#egg=kezmenu",
    ],
    entry_points={"console_scripts": ["fallingsky = fallingsky.main:play"]},
    url="https://github.com/a-tal/fallingsky",
    description="A game of falling blocks with RPG elements, uses pygame.",
    long_description=(
        "The Tragedy of the Falling Sky intends to be a RPG style block game. "
        "It is still in very early development though, so many features are "
        "missing. Also the artwork is very preliminary and will be improved "
        "upon. Feel free to contribute to the project on GitHub, all pull "
        "requests are welcomed."
    ),
    download_url="https://github.com/a-tal/fallingsky",
    tests_require=["pytest", "pytest-cov"],
    cmdclass={"test": PyTest},
    license="GPLv2",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Games/Entertainment",
        "Topic :: Games/Entertainment :: Arcade",
        "Topic :: Games/Entertainment :: Puzzle Games",
        "Topic :: Software Development :: Libraries :: pygame",
    ],
)



"""Score keeping class and helpers.

Usage:

>>> k = Keeper()
>>> k.game += 1000
>>> k.game_over()
>>> k.total
1000
>>> k.game
0
>>>

"""


class GameScore(object):
    def __init__(self, score=None, last_added=None):
        self._score = score or 0
        self._last_added = last_added or 0
        self._multiplier = 0
        super(GameScore, self).__init__()

    def __add__(self, other):
        return GameScore(self.get_score(), last_added=other)

    def get_score(self):
        """Calculates the current scores after applying multipliers."""

        return self._score + (self._last_added * (self._multiplier or 1))

    def multiply_last(self, multiplier):
        """Adds the multiplier to the last added score."""

        self._multiplier += multiplier

    def __int__(self):
        """Override to call this as int(GameScore)."""

        return self.get_score()

    def __str__(self):
        return str(self.get_score())

    def __repr__(self):
        return "<{}.{} object at {} with value {}>".format(
            __name__, self.__class__.__name__, hex(id(self)), self.get_score()
        )


class TotalScore(int):
    def __add__(self, other):
        return TotalScore(int(self) + int(other))


class Keeper(object):
    def __init__(self):
        self.game = GameScore(0)
        self.total = TotalScore(0)
        self.best = GameScore(0)

    def game_over(self):
        self.total += self.game
        self.best = max(self.best, self.game)
        self.game = GameScore(0)

    def resume(self, previous_total, previous_best):
        self.total = TotalScore(previous_total)
        self.best = GameScore(previous_best)


if __name__ == "__main__":
    k = Keeper()
    k.game += 1000
    print(k.game)
    k.game.multiply_last(2)
    print(k.game.get_score())
    print(repr(k.game))
